"""
Services.
"""
import logging
import requests
from urllib.parse import urlparse, parse_qs

from drf_requests_jwt import settings
from drf_requests_jwt.backends.utils import build_url

logger = logging.getLogger(__name__)


class HttpRequestService(object):
    obtain_jwt_allowed_fail_attempts = settings.DEFAULTS.get('OBTAIN_JWT_ALLOWED_FAIL_ATTEMPTS')
    cache_backend_class = settings.DEFAULTS.get('CACHE_BACKEND_CLASS')

    def __init__(self, params=None):
        super().__init__()

        self.cache_backend = self._get_cache_backend()

        self.headers = self._get_headers()
        self.url = self._get_url()

        self.params = params or {}
        self.params.update(self._get_params())

        self.session = requests.Session()

        self.obtain_jwt_fail_attempts = 0

    def _get_cache_backend(self):
        resolved_backend_class = settings.import_from_string(self.cache_backend_class)
        return resolved_backend_class(self._get_jwt_cache_key())

    def _get_base_url(self):
        raise NotImplementedError

    def _get_url_path(self):
        raise NotImplementedError

    def _get_url(self):
        return build_url(base_url=self._get_base_url(), path=self._get_url_path())

    def _get_jwt_login_url_path(self):
        raise NotImplementedError

    def _get_jwt_login_url(self):
        return build_url(base_url=self._get_base_url(), path=self._get_jwt_login_url_path())

    def _get_username(self):
        raise NotImplementedError

    def _get_password(self):
        raise NotImplementedError

    def _get_params(self):
        return {}

    def _get_headers(self):
        return {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Authorization': 'JWT {token}'.format(token=self._get_jwt_token_from_cache())
        }

    def get_results_from_all_pages(self):
        next_url = self.url
        result_list = []

        while True:
            url_parse = urlparse(next_url)

            self.params.update(parse_qs(url_parse.query))

            next_url = '{scheme}://{netloc}{path}'.format(
                scheme=url_parse.scheme, netloc=url_parse.netloc, path=url_parse.path
            )

            response = self.session.get(next_url, headers=self.headers, params=self.params)

            logger.debug('Request url: {} with params {}'.format(next_url, self.params))

            if response.status_code == 200:
                response_json = response.json()
                next_url = response_json.get('next')
                result_list.extend(response_json.get('results', []))
            elif response.status_code == 401:
                if self._should_update_authorization_header():
                    self.update_authorization_header()
                else:
                    break
            else:
                raise Exception('Wrong response status code: {code}, content: {content}'.format(
                    code=response.status_code,
                    content=response.content
                ))

            if not bool(next_url):
                break

        return result_list

    def write_results_from_all_pages_to_file(self, filename):
        results = self.get_results_from_all_pages()

        with open(filename, 'w') as output:
            json.dump(results, output)

    def update_authorization_header(self):
        token = self._get_jwt_token()
        self.headers['Authorization'] = 'JWT {token}'.format(token=token)

    def get_deserialized_data(self):
        raise NotImplementedError

    def _get_jwt_token(self):
        payload = {
            'username': self._get_username(),
            'password': self._get_password()
        }
        url = self._get_jwt_login_url()
        logger.debug('Request url: {}'.format(url))
        response = self.session.post(url, data=payload)

        if response.status_code == 200:
            response_dict = response.json()
            token = response_dict.get('token')
            self._set_jwt_token_to_cache(token)
            logger.debug('Received a fresh JWT token')
            return token
        else:
            self.obtain_jwt_fail_attempts += 1
            logger.warning('Attempt to get a JWT token failed')
            raise Exception('Wrong response status code: {code}, content: {content}'.format(
                code=response.status_code,
                content=response.content
            ))

    def _should_update_authorization_header(self):
        return self.obtain_jwt_fail_attempts <= self.obtain_jwt_allowed_fail_attempts

    def _set_jwt_token_to_cache(self, token):
        self.cache_backend.set_jwt(token)

    def _get_jwt_token_from_cache(self):
        return self.cache_backend.get_jwt()

    def _get_jwt_cache_key(self):
        return 'jwt-{url}-{username}'.format(url=self._get_base_url(), username=self._get_username())



from django.core.cache import cache

from drf_requests_jwt.backends.base import BaseBackend


class DjangoCacheBackend(BaseBackend):
    def get_jwt(self):
        return cache.get(self.key)

    def set_jwt(self, token):
        cache.set(self.key, token)



from unittest import TestCase

import mock

from drf_requests_jwt.services import HttpRequestService


class BaseHttpRequestServiceTestCase(TestCase):
    def _patch_requests(self):
        patcher = mock.patch('drf_requests_jwt.services.requests')
        self.requests_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def _patch_get_jwt_login_url_path(self):
        patcher = mock.patch('drf_requests_jwt.services.HttpRequestService._get_jwt_login_url_path')
        self.get_jwt_login_url_path_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def _patch_get_url_path(self):
        patcher = mock.patch('drf_requests_jwt.services.HttpRequestService._get_url_path')
        self.get_url_path_mock = patcher.start()
        self.get_url_path_mock.return_value = '/mocked/path/'
        self.addCleanup(patcher.stop)

    def _patch_get_password(self):
        patcher = mock.patch('drf_requests_jwt.services.HttpRequestService._get_password')
        self.get_password_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def _patch_get_username(self):
        patcher = mock.patch('drf_requests_jwt.services.HttpRequestService._get_username')
        self.get_username_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def _patch_get_base_url(self):
        patcher = mock.patch('drf_requests_jwt.services.HttpRequestService._get_base_url')
        self.get_base_url_mock = patcher.start()
        self.get_base_url_mock.return_value = 'http://base:1234'
        self.addCleanup(patcher.stop)

    def _patch_get_headers(self):
        patcher = mock.patch('drf_requests_jwt.services.HttpRequestService._get_headers')
        self.get_headers_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def _patch_get_cache_backend(self):
        patcher = mock.patch('drf_requests_jwt.services.HttpRequestService._get_cache_backend')
        self.get_cache_backend_mock = patcher.start()
        self.addCleanup(patcher.stop)


class HttpRequestServiceTestCase(BaseHttpRequestServiceTestCase):
    def setUp(self):
        super().setUp()

        self._patch_requests()
        self._patch_get_cache_backend()
        self._patch_get_headers()
        self._patch_get_base_url()
        self._patch_get_username()
        self._patch_get_password()
        self._patch_get_url_path()
        self._patch_get_jwt_login_url_path()

        self.session_mock = self.requests_mock.Session
        self.get_mock = self.session_mock.return_value.get
        self.post_mock = self.session_mock.return_value.post

    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_params')
    def test_init(self, get_params_mock):
        headers_mock = {'h1': 'hv1', 'h2': 'hv2'}
        url_mock = 'mock://host:1234/path/to/resource/'
        params_mock = {'p1': 'pv2', 'p2': 'pv2'}

        self.get_headers_mock.return_value = headers_mock
        self.get_base_url_mock.return_value = 'mock://host:1234'
        self.get_url_path_mock.return_value = 'path/to/resource/'
        get_params_mock.return_value = params_mock

        instance = HttpRequestService(params={'other_param': 'value'})

        self.assertDictEqual(instance.headers, headers_mock)
        self.assertEqual(instance.url, url_mock)
        self.assertDictEqual(instance.params, {'p1': 'pv2', 'p2': 'pv2', 'other_param': 'value'})
        self.assertEqual(instance.session, self.requests_mock.Session.return_value)

    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_base_url')
    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_url_path')
    def test_get_url(self, get_url_path_mock, get_base_url_mock):
        get_base_url_mock.return_value = 'http://base:1234'
        get_url_path_mock.return_value = 'path/to/resource/'

        instance = HttpRequestService()
        actual_result = instance._get_url()
        self.assertEqual(actual_result, 'http://base:1234/path/to/resource/')

    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_base_url')
    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_url_path')
    def test_get_url_with_trailing_slash(self, get_url_path_mock, get_base_url_mock):
        get_base_url_mock.return_value = 'http://base:1234/'
        get_url_path_mock.return_value = 'path/to/resource/'

        instance = HttpRequestService()
        actual_result = instance._get_url()
        self.assertEqual(actual_result, 'http://base:1234/path/to/resource/')

    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_base_url')
    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_jwt_login_url_path')
    def test_get_jwt_login_url(self, get_jwt_login_url_path_mock, get_base_url_mock):
        get_base_url_mock.return_value = 'http://base:1234'
        get_jwt_login_url_path_mock.return_value = 'path/to/resource/'

        instance = HttpRequestService()
        actual_result = instance._get_jwt_login_url()
        self.assertEqual(actual_result, 'http://base:1234/path/to/resource/')

    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_base_url')
    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_jwt_login_url_path')
    def test_get_jwt_login_url_with_trailing_slash(self, get_jwt_login_url_path_mock, get_base_url_mock):
        get_base_url_mock.return_value = 'http://base:1234/'
        get_jwt_login_url_path_mock.return_value = 'path/to/resource/'

        instance = HttpRequestService()
        actual_result = instance._get_jwt_login_url()
        self.assertEqual(actual_result, 'http://base:1234/path/to/resource/')

    def test_get_jwt_cache_key(self):
        self.get_base_url_mock.return_value = 'http://base:1234'
        self.get_username_mock.return_value = 'joe'

        instance = HttpRequestService()
        actual_result = instance._get_jwt_cache_key()
        self.assertEqual(actual_result, 'jwt-http://base:1234-joe')

    def test_get_params(self):
        instance = HttpRequestService()
        self.assertDictEqual(instance._get_params(), {})

    def test_get_results_from_all_pages_successful(self):
        headers_mock = mock.Mock()

        self.get_mock.return_value.status_code = 200
        self.get_mock.return_value.json.side_effect = [
            {'next': 'mock://host1/path1/?offset=2&limit=2', 'results': ['a', 'b']},
            {'next': 'mock://host2/path2/?offset=2&limit=2', 'results': ['x', 'y']},
            {'next': None, 'results': ['h']}
        ]

        self.get_base_url_mock.return_value = 'mock://host0'
        self.get_url_path_mock.return_value = 'path0/'
        params_mock = {'param1': ['1'], 'param2': ['2']}

        instance = HttpRequestService(params=params_mock)

        instance.headers = headers_mock

        actual_result = instance.get_results_from_all_pages()

        self.assertListEqual(actual_result, ['a', 'b', 'x', 'y', 'h'])

        self.get_mock.assert_any_call('mock://host0/path0/', headers=headers_mock, params=params_mock)
        params_mock.update({'offset': ['2'], 'limit': ['2']})

        self.get_mock.assert_any_call('mock://host1/path1/', headers=headers_mock, params=params_mock)
        self.get_mock.assert_any_call('mock://host2/path2/', headers=headers_mock, params=params_mock)

        self.assertEqual(self.get_mock.call_count, 3)

    @mock.patch('drf_requests_jwt.services.HttpRequestService._should_update_authorization_header')
    @mock.patch('drf_requests_jwt.services.HttpRequestService.update_authorization_header')
    def test_get_results_from_all_pages_unauthorized_should_update_header(self, update_mock, should_update_mock):
        should_update_mock.side_effect = [True, False]

        headers_mock = mock.Mock()
        params_mock = mock.Mock()

        self.get_mock.return_value.status_code = 401

        self.get_base_url_mock.return_value = 'mock://host0'
        self.get_url_path_mock.return_value = 'path0/'

        instance = HttpRequestService()

        instance.headers = headers_mock
        instance.params = params_mock

        actual_result = instance.get_results_from_all_pages()

        self.assertListEqual(actual_result, [])

        self.get_mock.assert_any_call('mock://host0/path0/', headers=headers_mock, params=params_mock)
        self.assertEqual(self.get_mock.call_count, 2)
        self.assertEqual(should_update_mock.call_count, 2)
        update_mock.assert_called_once_with()

    @mock.patch('drf_requests_jwt.services.HttpRequestService._should_update_authorization_header')
    @mock.patch('drf_requests_jwt.services.HttpRequestService.update_authorization_header')
    def test_get_results_from_all_pages_unauthorized_should_not_update_header(self, update_mock, should_update_mock):
        should_update_mock.return_value = False

        headers_mock = mock.Mock()
        params_mock = mock.Mock()

        self.get_mock.return_value.status_code = 401

        self.get_base_url_mock.return_value = 'mock://host0'
        self.get_url_path_mock.return_value = 'path0/'

        instance = HttpRequestService()

        instance.headers = headers_mock
        instance.params = params_mock

        actual_result = instance.get_results_from_all_pages()

        self.assertListEqual(actual_result, [])

        self.get_mock.assert_called_once_with('mock://host0/path0/', headers=headers_mock, params=params_mock)
        should_update_mock.assert_called_once_with()
        self.assertEqual(update_mock.call_count, 0)

    def test_get_results_from_all_pages_error(self):
        headers_mock = mock.Mock()
        params_mock = mock.Mock()

        self.get_mock.return_value.status_code = 404

        self.get_base_url_mock.return_value = 'mock://host0'
        self.get_url_path_mock.return_value = 'path0/'

        instance = HttpRequestService()

        instance.headers = headers_mock
        instance.params = params_mock

        self.assertRaises(Exception, instance.get_results_from_all_pages)

        self.get_mock.assert_called_once_with('mock://host0/path0/', headers=headers_mock, params=params_mock)

    def test_should_update_authorization_header_true(self):
        instance = HttpRequestService()

        instance.obtain_jwt_fail_attempts = 2
        instance.obtain_jwt_allowed_fail_attempts = 3

        self.assertTrue(instance._should_update_authorization_header())

        instance.obtain_jwt_fail_attempts = 3
        instance.obtain_jwt_allowed_fail_attempts = 3

        self.assertTrue(instance._should_update_authorization_header())

    def test_should_update_authorization_header_false(self):
        instance = HttpRequestService()

        instance.obtain_jwt_fail_attempts = 4
        instance.obtain_jwt_allowed_fail_attempts = 3

        self.assertFalse(instance._should_update_authorization_header())

    def test_set_jwt_token_to_cache(self):
        instance = HttpRequestService()
        instance._set_jwt_token_to_cache('token123')

        self.get_cache_backend_mock.return_value.set_jwt.assert_called_once_with('token123')

    def test_get_jwt_token_from_cache(self):
        expected_result = mock.Mock()
        self.get_cache_backend_mock.return_value.get_jwt.return_value = expected_result

        instance = HttpRequestService()

        actual_result = instance._get_jwt_token_from_cache()

        self.assertEqual(actual_result, expected_result)

        self.get_cache_backend_mock.return_value.get_jwt.assert_called_once_with()

    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_jwt_token')
    def test_update_authorization_header(self, get_jwt_token_mock):
        get_jwt_token_mock.return_value = 'updated-token12345'

        self.get_headers_mock.return_value = {}

        instance = HttpRequestService()

        instance.update_authorization_header()

        self.assertEqual(instance.headers['Authorization'], 'JWT updated-token12345')

    @mock.patch('drf_requests_jwt.services.HttpRequestService._set_jwt_token_to_cache')
    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_jwt_login_url')
    def test_get_jwt_token_success(self, get_jwt_login_url_mock, set_jwt_token_to_cache_mock):
        self.post_mock.return_value.status_code = 200
        self.post_mock.return_value.json.return_value = {'token': 'the-token12345'}

        get_jwt_login_url_mock.return_value = 'mock://host0:1234/path/to/jwt/login/'

        instance = HttpRequestService()

        actual_result = instance._get_jwt_token()

        self.assertEqual(actual_result, 'the-token12345')

        self.post_mock.assert_called_once_with('mock://host0:1234/path/to/jwt/login/', data={
            'username': self.get_username_mock.return_value,
            'password': self.get_password_mock.return_value,
        })

        set_jwt_token_to_cache_mock.assert_called_once_with('the-token12345')

    @mock.patch('drf_requests_jwt.services.HttpRequestService._set_jwt_token_to_cache')
    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_jwt_login_url')
    def test_get_jwt_token_fail(self, get_jwt_login_url_mock, set_jwt_token_to_cache_mock):
        self.post_mock.return_value.status_code = 400
        self.post_mock.return_value.json.return_value = {'token': 'the-token12345'}

        get_jwt_login_url_mock.return_value = 'mock://host0:1234/path/to/jwt/login/'

        instance = HttpRequestService()

        self.assertRaises(Exception, instance._get_jwt_token)

        self.post_mock.assert_called_once_with('mock://host0:1234/path/to/jwt/login/', data={
            'username': self.get_username_mock.return_value,
            'password': self.get_password_mock.return_value,
        })

        self.assertEqual(set_jwt_token_to_cache_mock.call_count, 0)


class HttpRequestHeadersTestCase(BaseHttpRequestServiceTestCase):
    def setUp(self):
        self._patch_get_base_url()
        self._patch_get_base_url()
        self._patch_get_username()
        self._patch_get_password()
        self._patch_get_url_path()

    @mock.patch('drf_requests_jwt.services.HttpRequestService._get_jwt_token_from_cache')
    def test_get_headers(self, get_jwt_token_from_cache_mock):
        get_jwt_token_from_cache_mock.return_value = 'jwt-username-token123456'
        instance = HttpRequestService()
        self.assertDictEqual(instance._get_headers(), {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Authorization': 'JWT jwt-username-token123456'
        })



from pathlib import Path

from drf_requests_jwt.backends.base import BaseBackend


class FileCacheBackend(BaseBackend):
    def __init__(self, key):
        super().__init__(key)

        self.file_path = '/tmp/{key}'.format(key=self.key)

    def get_jwt(self):
        if Path(self.file_path).is_file():
            with open(self.file_path, 'r') as f:
                return f.read()
        return None

    def set_jwt(self, token):
        with open(self.file_path, 'w') as f:
            f.write(token)



from unittest import TestCase

import mock

from drf_requests_jwt.deserializers import ObjectListDeserializerMixin


class ObjectListDeserializerMixinTestCase(TestCase):
    def test_get_deserialized_object_list_raise_exception(self):
        instance = ObjectListDeserializerMixin()
        self.assertRaises(Exception, instance.get_deserialized_object_list)

    def test_get_deserialized_object_list_all_valid(self):
        instance = ObjectListDeserializerMixin()
        instance.get_results_from_all_pages = mock.Mock(return_value=[{'k': 'o1'}, {'k': 'o2'}, {'k': 'o3'}])
        instance.serializer_class = mock.Mock()

        object1 = mock.Mock()
        object2 = mock.Mock()
        object3 = mock.Mock()

        serializer1 = mock.Mock(is_valid=mock.Mock(return_value=True), save=mock.Mock(return_value=object1))
        serializer2 = mock.Mock(is_valid=mock.Mock(return_value=True), save=mock.Mock(return_value=object2))
        serializer3 = mock.Mock(is_valid=mock.Mock(return_value=True), save=mock.Mock(return_value=object3))

        instance.serializer_class.side_effect = [serializer1, serializer2, serializer3]

        actual_result = instance.get_deserialized_object_list()

        self.assertListEqual(actual_result, [object1, object2, object3])

        instance.get_results_from_all_pages.assert_called_once_with()

        instance.serializer_class.assert_any_call(data={'k': 'o1'})
        instance.serializer_class.assert_any_call(data={'k': 'o2'})
        instance.serializer_class.assert_any_call(data={'k': 'o3'})

        self.assertEqual(instance.serializer_class.call_count, 3)

    def test_get_deserialized_object_list_one_invalid(self):
        instance = ObjectListDeserializerMixin()
        instance.get_results_from_all_pages = mock.Mock(return_value=[{'k': 'o1'}, {'k': 'o2'}, {'k': 'o3'}])
        instance.serializer_class = mock.Mock()

        object1 = mock.Mock()
        object2 = mock.Mock()
        object3 = mock.Mock()

        serializer1 = mock.Mock(is_valid=mock.Mock(return_value=True), save=mock.Mock(return_value=object1))
        serializer2 = mock.Mock(is_valid=mock.Mock(return_value=False), save=mock.Mock(return_value=object2))
        serializer3 = mock.Mock(is_valid=mock.Mock(return_value=True), save=mock.Mock(return_value=object3))

        instance.serializer_class.side_effect = [serializer1, serializer2, serializer3]

        actual_result = instance.get_deserialized_object_list()

        self.assertListEqual(actual_result, [object1, object3])

        instance.get_results_from_all_pages.assert_called_once_with()

        instance.serializer_class.assert_any_call(data={'k': 'o1'})
        instance.serializer_class.assert_any_call(data={'k': 'o2'})
        instance.serializer_class.assert_any_call(data={'k': 'o3'})

        self.assertEqual(instance.serializer_class.call_count, 3)



import urllib
from urllib.parse import urlparse


def build_url(base_url, path, args_dict=None):
    # Returns a list in the structure of urlparse.ParseResult
    url_parts = list(urlparse(base_url))
    url_parts[2] = path
    if args_dict is not None:
        url_parts[4] = urllib.parse.urlencode(args_dict)
    return urllib.parse.urlunparse(url_parts)



from slugify import slugify


class BaseBackend(object):
    def __init__(self, key):
        self.key = slugify(key)

    def get_jwt(self):
        raise NotImplementedError

    def set_jwt(self, token):
        raise NotImplementedError



from setuptools import setup, find_packages

requires = [
    'requests>=2.18.1',
    'python-slugify>=1.2.4'
]

setup(
    name='drf_requests_jwt',
    version='0.7',
    description='Django Rest Framework Requests with JWT support',
    long_description=open('README.rst').read(),
    url='https://github.com/sensidev/drf-requests-jwt',
    author='Sensidev',
    author_email='lucian.corduneanu@sensidev.com',
    license='MIT',
    packages=find_packages(exclude=["tests*"]),
    install_requires=requires,
    setup_requires=['pytest-runner', 'requests'],
    tests_require=['pytest', 'mock'],
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

)



"""
Deserializers.
"""
import logging

logger = logging.getLogger(__name__)


class ObjectListDeserializerMixin(object):
    serializer_class = None

    def get_deserialized_object_list(self):
        object_list = []

        if self.serializer_class is None:
            raise Exception('`serializer_class` is required')

        for object_json in self.get_results_from_all_pages():
            serializer = self.serializer_class(data=object_json)
            if serializer.is_valid():
                object_list.append(serializer.save())
            else:
                logger.debug("Serializer has errors {}", serializer.errors)

        return object_list
